#ifndef COMMON_H
#define COMMON_H

#include <cassert>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "elcore50-matrix-lib/dmainit.h"
#include "elcore50-matrix-lib/elcore50.h"

#define TIC_CNTR ((volatile uint32_t *)(E50BASER+0x280+0x10))
#define INSTR_CNTR ((volatile uint32_t *)(E50BASER+0x280+0x18))
#define PCU12_OFFSET 0x800
#define InvCtrl ((volatile uint32_t *)(E50BASER+PCU12_OFFSET+0x4))
#define Ctrl ((volatile uint32_t *)(E50BASER+PCU12_OFFSET+0x0))
#define L1DC_ctrl ((volatile uint32_t *)(E50BASER+0x34400))
#define L0DC_ctrl ((volatile uint32_t *)(E50BASER+0x36200))

extern int __local_mem;

enum Store_version
{
    STORE_NONE,
    STORE_RELU,
    STORE_RELU6,
};


static void change_L1DC_ctrl()
{
    *L0DC_ctrl &= 0x5;
    *L1DC_ctrl &= 0x24B44B41;
    *L1DC_ctrl &= 0xff83ffff; // <- 2385 (2212)
    *Ctrl = *Ctrl | 0x00000004; // <- 2324 (2212)
    *Ctrl = *Ctrl | 0x00008000; // <- 2387 (2212)
    *Ctrl = *Ctrl | 0x00030000; // <- 3042 (2212)
    *Ctrl = *Ctrl | 0x00040000; // <- 2387 (2212)
    *Ctrl = *Ctrl & 0xffe7ffff; // <- 2387 (2212)
    *Ctrl = *Ctrl | 0x00100000; // <- 2564 (2212)
}

static void flush_all_caches()
{
    asm volatile("mbar 0x0");
    asm volatile("sev0 18");
    asm volatile("sev0 19");
    *InvCtrl = 0xffff;
    asm volatile("wfe1 18");
    asm volatile("wfe1 19");
    asm volatile("sev0 18");
    asm volatile("sev0 19");
    asm volatile("mbar 0x0");
    change_L1DC_ctrl();
}

static void count_tics(int* tic_addr, int* instr_addr)
{
    *tic_addr = *TIC_CNTR;
    *instr_addr = *INSTR_CNTR;
}

typedef struct LocalMemBufFl16{
    uint16_t *buf_A[2], *buf_B[2], *buf_C[2];
    float *buf_init_vec;
    int buf_row0, buf_row1col0, buf_col1;
} LocalMemBufFl16;

typedef struct LocalMemBuf{
    float *buf_A[2], *buf_B[2], *buf_C[2];
    int buf_row0, buf_row1col0, buf_col1;
} LocalMemBuf;

typedef struct ConvLocalMemBufFl16{
    uint16_t *buf_A[2], *buf_B[2], *buf_C[2];
    uint16_t* ones;
    float *buf_init_vec;
    int buf_srcH, buf_srcW, buf_srcC, buf_dstH, buf_dstW, buf_dstC, offset_A;
    int tile_padX, tile_padY, tile_padW, tile_padH;
} ConvLocalMemBufFl16;

typedef struct MatMulDmaChains4{
    VDMAChain* chain_A;
    VDMAChain* chain_B;
    VDMAChain* chain_ld_C;
    VDMAChain* chain_st_C;
    VDMAChain* chain_init_vec;
    int len_chain_A;
    int len_chain_B;
    int len_chain_ld_C;
    int len_chain_st_C;
}MatMulDmaChains4;

#endif
