#include <cassert>
#include <stdio.h>
#include <string.h>

#include "elcore50-matrix-lib/common.h"

//#define DEBUG
//#define NO_CALC
//#define NO_DMA
//#define MANUAL_TILE_SIZE

typedef struct MatMulDmaChains{
    VDMAChain* chain;
    VDMAChain* chain_B;
    VDMAChain* chain_st;
}MatMulDmaChains;

typedef struct LastStore{
    float *src;
    float* dst;
    int width;
    int height;
    int offset_src;
    int offset_dst;
    int buf_num;

} LastStore;

void size_selector(int M, int K, int N, int& buf_M, int& buf_K, int& buf_N,
                   int count_buf_A, int count_buf_B, int count_buf_C);

void size_selector_fl32_offsets_K(int M, int K, int N, int& buf_M, int& buf_K, int& buf_N,
                                  int count_buf_A, int count_buf_B, int count_buf_C,
                                  int& offsetA, int& offsetB);

extern "C" void mm_v0_vliw_1_sub_matrix_pre_load_real_out_offset(float *in1, int row, int col0row1, float *in2, int col1,
                                                                float *out, int x_offset, int y_offset, int real_row, int real_col,
                                                                int* tics, int* instr, int offsetA, int offsetB);
extern "C" void mm_v0_vliw_1_sub_matrix_pre_load_real_out_offset_local_mem(float *in1, int row, int col0row1, float *in2, int col1,
                                                                float *out, int x_offset, int y_offset, int real_row, int real_col,
                                                                int* tics, int* instr, int offsetA, int offsetB);
extern "C" void mm_16x64(float *in1, int row, int col0row1, float *in2, int col1, float *out, int flag, int real_col,
                         int* tics, int* instr, int offsetA, int offsetB);
extern "C" void mm_16x64_local_mem(float *in1, int row, int col0row1, float *in2, int col1, float *out, int flag, int real_col,
                                   int* tics, int* instr, int offsetA, int offsetB);
extern "C" void mm_32x32_local_mem(float *in1, int row, int col0row1, float *in2, int col1, float *out, int flag, int real_col,
                                   int* tics, int* instr, int offsetA, int offsetB);

void init_chain_dma_two_buf_ver_nkm(float* src0, int row0, int row1col0, float* src1, int col1,
                                    float* dst, int& offset_A, int& offset_B,
                                    LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_two_buf_ver_nkm(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                                 LocalMemBuf *buffer, MatMulDmaChains4* chains);

void init_chain_two_buf_ver_nkm_v2(float* src0, int row0, int row1col0, float* src1, int col1,
                                   float* dst, int& offset_A, int& offset_B,
                                   LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_two_buf_ver_nkm_v2(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                                    LocalMemBuf *buffer, MatMulDmaChains4* chains);

void init_chain_one_buf_ver_nkm(float* src0, int row0, int row1col0, float* src1, int col1,
                                float* dst,  int& offset_A, int& offset_B,
                                LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_one_buf_ver_nkm(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                                 LocalMemBuf *buffer, MatMulDmaChains4* chains);

void init_chain_two_buf_v2(float* src0, int row0, int row1col0, float* src1, int col1,
                           float* dst,  int& offset_A, int& offset_B,
                           LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_two_buf_v2(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                            LocalMemBuf *buffer, MatMulDmaChains4* chains);

void init_chain_two_buf_v2_16x64(float* src0, int row0, int row1col0, float* src1, int col1,
                                 float* dst, int& offset_A, int& offset_B,
                                 LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_two_buf_v2_16x64(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                                  LocalMemBuf *buffer, MatMulDmaChains4* chains);

void init_chain_two_buf_v2_16x64_optional(float* src0, int row0, int row1col0, float* src1, int col1,
                                          float* dst, int& offset_A, int& offset_B,
                                          LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_two_buf_v2_16x64_optional(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                                           LocalMemBuf *buffer, MatMulDmaChains4* chains);

void init_chain_two_buf_v3_16x64_optional(float* src0, int row0, int row1col0, float* src1, int col1,
                                          float* dst, int& offset_A, int& offset_B,
                                          LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_two_buf_v3_16x64_optional(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                                           LocalMemBuf *buffer, MatMulDmaChains4* chains);

void init_chain_dma_two_buf_ver_nkm_ver_1C(float* src0, int row0, int row1col0, float* src1, int col1,
                                           float* dst, int& offset_A, int& offset_B,
                                           LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_two_buf_ver_nkm_ver_1C(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                                        LocalMemBuf *buffer, MatMulDmaChains4* chains);

void init_chain_two_buf_4_chains(float* src0, int row0, int row1col0, float* src1, int col1,
                                 float* dst, int& offset_A, int& offset_B,
                                 LocalMemBuf* buffer, MatMulDmaChains4* chains);
void mat_mul_dma_two_buf_4_chains(float* src0, int row0, int row1col0, float* src1, int col1, float* dst, int offset_A, int offset_B,
                                  LocalMemBuf *buffer, MatMulDmaChains4* chains);
